----------------------------------------------------------------------------------
-- Company: Boston University Electronic Design Facility (edf.bu.edu)
-- Engineer: Christopher Woodall (cwoodall at bu.edu)
-- 
-- Create Date:    12:13:43 05/22/2012 
-- Design Name:    Open Ethernet Mac 10/100 (Full Duplex)
-- Module Name:    ethernet_mac - Behavioral 
-- Project Name: 
-- Target Devices: Xilinx Spartan 6 chips 
-- Tool versions: ISE 13.3
-- Description: OEMAC is an open ethernet mac that is designed to interface for
--              10/100Mbs Full Duplex MII Ethernet PHYs. Its output interface
--              is very similar to the Xilinx TEMAC, in terms that it outputs 1
--              byte at a time as the packets come in and returns a packet good
--              or packet bad message dep
--
-- Dependencies: IEEE.STD_LOGIC_1164, crc_check.vhd
--
-- Version: 0.2 (unstable)
-- Additional Comments:
--   * Rx is now more reliable. Tx has been reliable so far.
--
----------------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.all;

entity ethernet_mac is
  Port(
    -- Interface with the PHY chip
    rxd   : in std_logic_vector(3 downto 0);
    rxdv  : in std_logic;
    rxclk : in std_logic;
    rxer  : in std_logic;

    txd   : out std_logic_vector(3 downto 0);
    txen  : out std_logic;
    txer  : out std_logic;
    txclk : in  std_logic;

    crs  : in    std_logic;
    col  : in    std_logic;
    mdio : inout std_logic;
    mdc  : out   std_logic;  -- Not 100% if this should be in or out.

    -- Interface with the other FPGA modules
--    led              : out    std_logic;
    rx_mac_data      : out std_logic_vector (7 downto 0);
    rx_mac_clk       : in    std_logic;
    rx_mac_valid     : out    std_logic;
--    rx_mac_fcs       : buffer std_logic_vector (31 downto 0);
--    rx_mac_crc       : buffer std_logic_vector (31 downto 0);
    rx_mac_goodframe : out    std_logic;
    rx_mac_badframe  : out    std_logic;
    rx_mac_clk_en : in std_logic;
  
    tx_mac_clk   : in std_logic;
    tx_mac_clk_en : in std_logic;

    tx_mac_valid : in  std_logic;
    tx_mac_ack   : out std_logic;
    tx_mac_data  : in  std_logic_vector(7 downto 0);

    rst_n    : in std_logic;
    enable_n : in std_logic
    );      
end ethernet_mac;

architecture Behavioral of ethernet_mac is
  signal tx_mac_valid_prev : std_logic;  -- added
  
-- RX Signals
  -- Previous rxd nibble from the PHY chip.
  signal rxd_prev        : std_logic_vector (3 downto 0);
  -- A buffer which contains the received byte, being sent out.
  signal rx_mac_data_buf : std_logic_vector (7 downto 0);
  -- Shift Left Register. Shifts by bytes.
  signal rx_mac_data_ls  : std_logic_vector (31 downto 0);
  -- MAC is enabled. Internal State. Determined by rxdv
  signal rx_mac_en       : std_logic                     := '0';
  -- stores rx_mac_valid output in a shift left register.
  -- rx_mac_valid_ls(4) is the same as rx_mac_valid
  signal rx_mac_valid_ls : std_logic_vector (4 downto 0) := "00000";
  -- Not really a counter anymore, keeps track of the
  -- nibbles and there byte locations from the PHY chip.
  signal rx_counter      : std_logic                     := '0';
  signal rx_mac_eof      : std_logic                     := '0';
  --signal rx_mac_eof_p : std_logic := '0';
  component crc_check
    port (
      clk      : in std_logic;
      rst_n    : in std_logic;
      enable_n : in std_logic;

      data : in std_logic_vector(7 downto 0);

      crc_o : out std_logic_vector(31 downto 0)
      );
  end component;

  -----------------------------------------------------------------------------
  -- RX Signals
  -----------------------------------------------------------------------------

--  signal rx_mac_crc : std_logic_vector (31 downto 0);
--  signal rx_mac_fcs : std_logic_vector (31 downto 0);
  signal rx_mac_valid_buf : std_logic                      := '0';
  signal rx_mac_crc_rst_n : std_logic                      := '0';
  signal rx_mac_crc_buf   : std_logic_vector (31 downto 0) := x"00000000";
--  signal rx_mac_clk     : std_logic                      := '0';
  signal rx_mac_fcs : std_logic_vector (31 downto 0);
  signal rx_mac_crc : std_logic_vector (31 downto 0);  
  signal rx_mac_data_o_buf : std_logic_vector (7 downto 0);
  signal rx_data_rdy_buf      : std_logic;
  signal rx_mac_goodframe_buf : std_logic;
  signal rx_mac_badframe_buf  : std_logic;
  signal rx_mac_phase         : std_logic;
  signal rx_mac_almost_eof    : std_logic;

  -----------------------------------------------------------------------------
  -- TX Signals
  -----------------------------------------------------------------------------
--  signal tx_mac_clk     : std_logic                      := '0';
  signal tx_mac_send      : std_logic_vector (7 downto 0)  := x"00";
  signal byte_count       : integer range 0 to 1514;
  signal tx_mac_crc_rst_n : std_logic                      := '0';
  signal tx_mac_crc_buf   : std_logic_vector (31 downto 0) := x"00000000";
  signal PREAMBLE         : std_logic_vector (63 downto 0) := x"55555555555555D5";
  signal preamble_sent    : std_logic                      := '0';
  signal tx_mac_dval      : std_logic                      := '0';
  signal tx_mac_crc       : std_logic_vector (31 downto 0) := x"00000000";
  signal tx_mac_crc_ctr   : integer range 0 to 3           := 0;
begin
  -----------------------------------------------------------------------------
  -- RX FABRIC                                                               --
  -----------------------------------------------------------------------------
  -- Generate appropriate Clock Enable, that is expected. IPBus uses it like a
  -- clock, as will we. It is buffered, which is generally bad practice, but..

  mdc  <= '0';
  mdio <= crs and col;

  rx_mac_data <= rx_mac_data_o_buf;
  -- It should work.


--  BUFIO2(rx_mac_clk, rx_mac_clk);
--  rx_mac_clk <= rx_mac_clk;
  --gen_rx_mac_clk : process (rxclk)
  --begin
  --  if rising_edge(rxclk) then
  --    if rst_n = '0' then
  --      rx_mac_clk <= '0';
  --    else
  --      rx_mac_clk <= not rx_mac_clk;
  --    end if;
  --  end if;
  --end process;

  CRCRx001 : crc_check port map (
    clk      => rx_mac_clk,
    rst_n    => rx_mac_crc_rst_n,
    enable_n => '0',

    data  => rx_mac_data_o_buf,
--    data => rx_mac_data_ls(23 downto 16),
    crc_o => rx_mac_crc_buf
    );

--  rx_mac_phase <= '0';
  --rx_mac_phase <= '1';
  rx_phase_adjust : process(rxclk)
  begin
    if rising_edge(rxclk) then
 --    if rst_n = '0' then
--        rx_mac_phase <= '0';
--      else
        if (rx_counter = '1') then
--          rx_mac_phase <= rx_mac_clk_en;
          if rx_mac_clk_en = '1' then
            rx_mac_phase <= '1';
          else
            rx_mac_phase <= '0';
          end if;
        end if;
  --    end if;
    end if;
  end process;

  rx_mac_crc_rst_n <= rx_mac_valid_buf or rx_mac_eof;
  rx_mac_valid     <= rx_mac_valid_buf;
  receive_data : process(rxclk)
  begin
    if falling_edge(rxclk) then
      if rst_n = '0' then
        rx_counter      <= '0';
        rx_mac_en       <= '0';
        rx_mac_valid_ls <= "00000";
      else
        -- Detect if there has been a change of state in the rx data valid (RxDv) 
        -- signal from the PHY chip. If there has been raise the frame_start flag
        -- and wait for teh data to become valid to raise the rx_mac_valid flag.
        if (rxdv = '1' and rxer = '0') then
          -- just started sending a message assume preamble. Once preamble received, set
          -- rx_mac_valid_b to '1' since a valid message frame has begun.
          if rx_mac_en = '0' then
            if rxd = x"D" then
              rx_mac_en  <= '1';
              rx_counter <= '0';
            end if;
          else
            if rx_counter = '1' then
              -- Left shift in the new data byte rxd & rxd_prev outputting the top byte to 
              -- rx_mac_data (the output of the mac). Basically, the output is delayed
              -- by four such that the FCS (Frame Check Sequence) part of the ethernet
              -- frame should be left inside rx_mac_data_ls and can be checked directly.
              rx_mac_data_buf <= rx_mac_data_ls(31 downto 24);
              rx_mac_data_ls  <= rx_mac_data_ls(23 downto 0) & rxd & rxd_prev;
              -- that data is ready for the checkings (probably should be changed to rx_mac_clk)
              rx_counter      <= '0';
              rx_mac_valid_ls <= rx_mac_valid_ls(3 downto 0) & '1';
            else
              rx_counter <= '1';
              rxd_prev   <= rxd;
            end if;
          end if;
          
        else
          rx_mac_en       <= '0';
          rx_counter      <= '0';
          rx_mac_valid_ls <= "00000";
        end if;
      end if;
    end if;
  end process;

  rx_mac_goodframe <= rx_mac_goodframe_buf;
  rx_mac_badframe  <= rx_mac_badframe_buf;

  rx_output : process (rx_mac_clk)
  begin
    if falling_edge(rx_mac_clk) then
      if rst_n = '0' then
        rx_mac_goodframe_buf <= '0';
        rx_mac_badframe_buf  <= '0';
        rx_mac_valid_buf     <= '0';
        rx_mac_eof           <= '0';
        rx_mac_data_o_buf          <= x"00";
        rx_mac_crc           <= x"00000000";
        rx_mac_almost_eof    <= '0';
      else
        if rx_mac_eof = '1' then
          rx_mac_eof <= '0';
          rx_mac_goodframe_buf <= '0';
          rx_mac_badframe_buf <= '0';

          rx_mac_almost_eof <= '0';
        elsif rx_mac_valid_ls(4) = '1' then
          rx_mac_data_o_buf       <= rx_mac_data_buf;
          rx_mac_valid_buf  <= rx_mac_valid_ls(4);
          rx_mac_almost_eof <= '0';
          rx_mac_badframe_buf  <= '0';
          rx_mac_goodframe_buf <= '0';
        elsif rx_mac_valid_ls(4) = '0' and rx_mac_valid_buf = '1' then
          if rx_mac_phase = '1'  and rx_mac_almost_eof = '0' then
            rx_mac_almost_eof <= '1';
            rx_mac_badframe_buf  <= '0';
            rx_mac_goodframe_buf <= '0';
            rx_mac_data_o_buf       <= rx_mac_data_buf;
            if rx_mac_crc_buf = rx_mac_fcs then
              rx_mac_goodframe_buf <= '1';
              rx_mac_badframe_buf  <= '0';
            else
              rx_mac_badframe_buf  <= '1';
              rx_mac_goodframe_buf <= '0';
            end if;

          else
            rx_mac_almost_eof <= '0';
            rx_mac_valid_buf  <= rx_mac_valid_ls(4);
            rx_mac_eof        <= '1';

            -- Check CRC we calculated against the CRC sent (FCS)

            --!START Problem Area
            -- New firmware with crc checking seems to work, be cautious.
            -- Traditionally this has been a vector for errors
            if rx_mac_almost_eof = '0' then
            if rx_mac_crc_buf = rx_mac_fcs then
              rx_mac_goodframe_buf <= '1';
              rx_mac_badframe_buf  <= '0';
            else
              rx_mac_badframe_buf  <= '1';
              rx_mac_goodframe_buf <= '0';
            end if;
            end if;
            -- Terrible replacement, ignores CRC check. Not a good idea
            --rx_mac_goodframe_buf <= '1'; 
            --!END Problem Area

            rx_mac_data_o_buf       <= rx_mac_data_buf;
            rx_mac_crc        <= rx_mac_crc_buf;
          end if;
        else
          rx_mac_almost_eof    <= '0';
          rx_mac_valid_buf     <= rx_mac_valid_ls(4);
          rx_mac_badframe_buf  <= '0';
          rx_mac_goodframe_buf <= '0';
        end if;
      end if;
    end if;
  end process;
  rx_mac_fcs        <= rx_mac_data_ls(7 downto 0) & rx_mac_data_ls(15 downto 8) & rx_mac_data_ls(23 downto 16) & rx_mac_data_ls(31 downto 24);

  -----------------------------------------------------------------------------
  -- TX FABRIC                                                               -- 
  -----------------------------------------------------------------------------
  -- Generate appropriate Clock Enable, that is expected. IPBus uses it like a
  -- clock, as will we. It is buffered, which is generally bad practice, but...
  -- It should work.
--  tx_mac_clk <= tx_mac_clk;
  --gen_tx_mac_clk : process (txclk)
  --begin
  --  if rising_edge(txclk) then
  --    tx_mac_clk <= not tx_mac_clk;
  --  end if;
  --end process;  -- gen_tx_mac_clk

  -- TX CRC
  CRCTx001 : crc_check port map (
    clk      => tx_mac_clk,
    rst_n    => tx_mac_crc_rst_n,
    enable_n => '0',

    data  => tx_mac_send,
    crc_o => tx_mac_crc_buf
    );
  
  tx_valid_sampler : process (tx_mac_clk)
  begin
    if falling_edge(tx_mac_clk) then
      tx_mac_valid_prev <= tx_mac_valid;
    end if;
  end process;
  
  tx_client_interface : process (tx_mac_clk)
    -- Minimum message length is 46 minimum payload + 6 byte MAC dest + 6 byte
    -- MAC source + 2 byte Ethernet Type/Length. Zero pad to the minimum length
    constant MIN_LENGTH : integer := 60;
    -- Maximum message length is 1500 byte max paylod + 6 byte MAC dest + 6
    -- byte MAC source + 2 byte Ethernet type/length.
    constant MAX_LENGTH : integer := 1514;
  begin

    if falling_edge(tx_mac_clk) then
      if rst_n = '0' then
        tx_mac_ack       <= '0';
        preamble_sent    <= '0';
        preamble         <= x"55555555555555D5";
        tx_mac_send      <= x"00";
        byte_count       <= 0;
        tx_mac_crc_rst_n <= '0';
        tx_mac_dval      <= '0';
        tx_mac_crc_ctr   <= 0;
      else        
--        tx_mac_valid_prev <= tx_mac_valid;  -- Added
        
        -- If tx_mac_valid is raised, take the opportunity to send the
        -- preamble, then acknowledge that a byte was received and send that
        -- byte the CRC. Function as normal until tx_mac_valid is dropped.
        if tx_mac_valid = '1' and tx_mac_valid_prev = '1' then -- added
          if byte_count = 0 then
            tx_mac_dval <= '1';
            -- Send the preamble
            if preamble_sent = '0' then
              tx_mac_send <= PREAMBLE(63 downto 56);
              PREAMBLE    <= PREAMBLE(55 downto 0) & PREAMBLE(63 downto 56);
              if (PREAMBLE(63 downto 56) = x"D5") then
                preamble_sent <= '1';
              end if;
              -- Send the data and then acknowledge the first byte. Also, enable
              -- the TxCRC001 part.
            else
              tx_mac_send      <= tx_mac_data;
              tx_mac_ack       <= '1';
              tx_mac_crc_rst_n <= '1';
              byte_count       <= 1;
            end if;
          else
            preamble_sent <= '0';
            -- Send the rest of the data.
            tx_mac_ack    <= '0';
            tx_mac_send   <= tx_mac_data;
            byte_count    <= byte_count + 1;
      
          end if;
        -- When tx_mac_valid is dropped, pad to minimum length and then 
        -- finish the crc calculation and send the FCS.
        else
          preamble_sent <= '0';
          tx_mac_ack <= '0';
          if tx_mac_dval = '1' then
            -- Add padding.
            if (byte_count <= MIN_LENGTH) then
              byte_count  <= byte_count + 1;
              tx_mac_send <= x"00";
            elsif (byte_count > MIN_LENGTH) then
              if (tx_mac_crc_rst_n = '1') then
                -- Once we are above the MIN_LENGTH activate tx_mac_crc_rst_n
                -- Also take a capture of the tx_mac_crc_buf and output the first
                -- byte of the crc.
                tx_mac_crc       <= x"00" & tx_mac_crc_buf(31 downto 8);
                tx_mac_send      <= tx_mac_crc_buf(7 downto 0);
                tx_mac_crc_rst_n <= '0';
                tx_mac_crc_ctr   <= 0;
              else
                -- Send the remaining 3 crc bytes.
                tx_mac_crc  <= x"00" & tx_mac_crc(31 downto 8);
                tx_mac_send <= tx_mac_crc(7 downto 0);
                if (tx_mac_crc_ctr = 3) then
                  tx_mac_dval <= '0';
                else
                  tx_mac_crc_ctr <= tx_mac_crc_ctr + 1;
                end if;
              end if;
            end if;
          end if;
        end if;
      end if;
    end if;
  end process;

  transmit_data : process(txclk)
  begin
    if falling_edge(txclk) then
      if rst_n = '0' then
        txd  <= "0000";
        txen <= '0';
        txer <= '0';
      else
        if tx_mac_dval = '1' then
          txen <= '1';
          -- Send the messages least significant nibble first.
          if tx_mac_clk_en = '1' then
            txd <= tx_mac_send (3 downto 0);
          else
            txd <= tx_mac_send (7 downto 4);
          end if;
        else
          txen <= '0';
          txd  <= "0000";
          txer <= '0';
        end if;
      end if;
    end if;
  end process;
end Behavioral;
