-- The ipbus slaves live in this entity - modify according to requirements
--
-- Ports can be added to give ipbus slaves access to the chip top level.
--
-- Dave Newbold, February 2011

library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.NUMERIC_STD.ALL;
use work.ipbus.ALL;
use work.bus_arb_decl.all;
use work.ttt_funcs.all;
entity slaves is port(
  ipb_clk, rst : in STD_LOGIC;
  ipb_in : in ipb_wbus;
  ipb_out : buffer ipb_rbus;

  -- Top level ports from here
  --gpio : out STD_LOGIC_VECTOR(3 downto 0);
  parm : buffer std_logic_vector(255 downto 0);
  ar : buffer std_logic_vector(7 downto 0);
  cr : buffer std_logic_vector(7 downto 0);
  dr : in std_logic_vector(7 downto 0);
  sr : in std_logic_vector(7 downto 0);
  l1a_test : inout std_logic_vector(7 downto 0);
  oob_moti: out trans_moti;
  oob_tomi: in trans_tomi;
  flash_wdata : in std_logic_vector(31 downto 0);
  flash_rdata : in std_logic_vector (31 downto 0);
  flash_busy : in std_logic;
  ttt_ctrl : buffer std_logic_vector(31 downto 0);
  ttt_action : buffer std_logic_vector(31 downto 0);
  ttt_action_ack : in std_logic_vector(31 downto 0);
  ttt_bcast_cmd : buffer std_logic_vector(31 downto 0);
  chA_bcast_err_counter : in std_logic_vector(31 downto 0);
  chB_bcast_err_counter : in std_logic_vector(31 downto 0);
  eth_frame_err : in std_logic_vector(31 downto 0);

  MacSel : in std_logic_vector(3 downto 0)
  );
end slaves;

architecture rtl of slaves is
  constant NSLV: positive := 1;
  signal ipbw: ipb_wbus_array(NSLV-1 downto 0);
  signal ipbr, ipbr_d: ipb_rbus_array(NSLV-1 downto 0);
  signal rdata : std_logic_vector (31 downto 0);
  signal TTTStatusReg : std_logic_vector(31 downto 0);
  signal ttt_ar : std_logic_vector (31 downto 0);
begin
  updateStatus : process (flash_busy)
  begin
    TTTStatusReg <= x"0000000" & '0' & '0' & '0' & flash_busy;
  end process;
  
  writeIPBus : process (ipb_in, flash_wdata, flash_rdata, rdata)
  begin
    if (ipb_in.ipb_addr(12) = '1') then
      if(ipb_in.ipb_addr(7) = '0')then
        ipb_out.ipb_rdata <= flash_wdata;
      else
        ipb_out.ipb_rdata <= flash_rdata;
      end if;
    else
      ipb_out.ipb_rdata <= rdata;
    end if;
  end process;

  handleIPBusInteractions: process (ipb_clk, rst)
  begin
    -- Uses ttt_ipb_rw_reg to handle IPBus to register transactions. code can
    -- be found in ttt_funcs.vhd
    
--    if rising_edge(ipb_clk) then
      -- Note the following formatting for entries into the address table.
      -- > -- RegName ([R]eadPermision/[W]ritePermission)
      -- > --       Any Description Stuff
      -- > when x"ADDRESS" => stuff;
      --
      -- The purpose of this formatting is to aid in addressTable generation with
      -- consistent names and descriptions as well as R, W permissions. Here is
      -- an example of the formatting in use for register ExampleReg at address
      -- 0x00000001 , with Read  only permissions and no special mask (just uses
      -- 0xffffffff as its mask).
      --
      -- > -- ExampleReg (R)
      -- > --        I am an example register! Please read me!
      -- > when x"00000001" => ttt_ipb_read_const(ipb_in, ipb_out, x"DEADBEEF");
      --
      -- To indicate the start of an address table use:
      -- > --!START_ADDRESS_TABLE
      --
      -- To indicate the end of an address table use
      -- > --!END_ADDRESS_TABLE
      --
      -- Of course you can always manually maintain an address table. Results
      -- should be checked by the human eye. Script can be found in:
      -- PROJECTBASE/src/script/VHDL2addrTable/vhdl2addrTable.py

    -- Generate flash part of address table
    if rst = '1' then
      ttt_ctrl      <= x"00000000";
      ttt_action    <= x"00000000";
      ttt_bcast_cmd(7 downto 0) <= x"00";
    elsif rising_edge(ipb_clk) then
      if (ttt_action_ack(29) = '1' and ttt_action(29) = '1') then
          ttt_action(29) <= '0';
      end if;

      if (ttt_action_ack(1) = '1') then
          ttt_action(1) <= '0';
      end if;
      
      case ipb_in.ipb_addr(11 downto 0) is
        --!START_ADDRESS_TABLE
        -- FirmwareVersionID (R) and Serial Number
        --      Format: YYYmddSS
        --              SSS is the Serial Number
        --              R is the revision Letter (ASCII code)
        --              YYYmdd is the version number ([Y]ear [m]onth [d]ay)
        when x"000" => ttt_ipb_read_const(rdata, x"7DD" & x"6" & x"07" & x"0" & MacSel);
               
        -- TTTStatusReg (R)
        --      Reports the status of various subsystems.
        --      bit 0 -> flash_busy
        when x"001" => ttt_ipb_read_reg( rdata, TTTStatusReg);
        -- TTTControlReg (R/W)
        when x"002" => ttt_ipb_rw_reg(ipb_in, rdata, ttt_ctrl, 0, 32, '1');
        -- TTTActionReg (R/W)
        when x"003" => ttt_ipb_rw_reg(ipb_in, rdata, ttt_action, 0, 32, '1');
        -- TTTBroadcastCmd (R/W)
        when x"004" => ttt_ipb_rw_reg(ipb_in, rdata, ttt_bcast_cmd, 0, 32, '1');

        when x"005" => ttt_ipb_read_reg(rdata, chA_bcast_err_counter);
        when x"006" => ttt_ipb_read_reg(rdata, chB_bcast_err_counter);
        when x"007" => ttt_ipb_read_reg(rdata, eth_frame_err);
                       
        -- TTSOrbitLength (R/W)
        when x"100" => ttt_ipb_rw_reg(ipb_in, rdata, parm,  0, 12, '1');
        -- TTSTrigStart (R/W)            
        when x"101" => ttt_ipb_rw_reg(ipb_in, rdata, parm, 16, 12, '1');
        -- TTSOrbitMax (R/W)             
        when x"102" => ttt_ipb_rw_reg(ipb_in, rdata, parm, 32, 16, '1');
        -- TTSL1ASpacing (R/W)           
        when x"103" => ttt_ipb_rw_reg(ipb_in, rdata, parm, 48, 12, '1');
        -- TTSL1APerOrbit (R/W)          
        when x"104" => ttt_ipb_rw_reg(ipb_in, rdata, parm, 64, 8, '1');
        -- TTSRepeatPeriod (R/W)         
        when x"105" => ttt_ipb_rw_reg(ipb_in, rdata, parm, 72, 16, '1');
        -- TTSFifoByte (R/W)             
        when x"106" => ttt_ipb_rw_reg(ipb_in, rdata, parm, 88, 4, '1');
        -- TTSCounterByte (R/W)          
        when x"107" => ttt_ipb_rw_reg(ipb_in, rdata, parm, 96, 2, '1');
        -- TTSRateSet (R/W)              
        when x"108" => ttt_ipb_rw_reg(ipb_in, rdata, parm, 104, 16, '1');
        -- TTSRule1 (R/W)
        when x"109" => ttt_ipb_rw_reg(ipb_in, rdata, parm, 16*8, 8, '1');
        -- TTSRule2 (R/W)                
        when x"10a" => ttt_ipb_rw_reg(ipb_in, rdata, parm, 17*8, 8, '1');
        -- TTSRule3 (R/W)                
        when x"10b" => ttt_ipb_rw_reg(ipb_in, rdata, parm, 18*8, 8, '1');
                       -- TTSRule4 (R/W)                
        when x"10c" => ttt_ipb_rw_reg(ipb_in, rdata, parm, 19*8, 8, '1');
                       -- TTSClkSel (R/W)               
        when x"10d" => ttt_ipb_rw_reg(ipb_in, rdata, parm, 20*8, 2, '1');
                       -- TTSSetBpDelay (R/W)           
        when x"10e" => ttt_ipb_rw_reg(ipb_in, rdata, parm, 21*8, 8, '1');
        -- TTSBpSampleMask (R/W)         
        when x"10f" => ttt_ipb_rw_reg(ipb_in, rdata, parm, 22*8, 8, '1');
        -- TTSTtcBxNum (R/W)             
        when x"110" => ttt_ipb_rw_reg(ipb_in, rdata, parm, 23*8, 12, '1');
        -- TTSTtcCmdPrescale (R/W)       
        when x"111" => ttt_ipb_rw_reg(ipb_in, rdata, parm, 25*8, 16, '1');
        -- TTSControlReg (R/W)
        when x"112" => ttt_ipb_rw_reg(ipb_in, rdata, cr, 0, 8, '1');
        -- TTSStatusReg (R)
        when x"113" => ttt_ipb_read_reg( rdata, sr);
        -- TTSActionReg (R/W)
        when x"114" => ttt_ipb_rw_reg(ipb_in, rdata, ar, 0, 8, '1');
        -- TTSDataReg (R)
        when x"115" => ttt_ipb_read_reg(rdata, dr);

                       
        when others => ttt_ipb_read_const( rdata, x"00000000");
        --!END_ADDRESS_TABLE
      end case;     
      ipb_out.ipb_ack <= ipb_in.ipb_strobe;
      ipb_out.ipb_err <= '0';
    end if;
  end process;
end rtl;


