---------------------------------------------------------------------------------
-- Company: Boston University EDF
-- Engineer: Christopher Woodall
-- 
-- Create Date:    17:10:03 06/06/2012 
-- Design Name: TTC Decoder DDR
-- Module Name:    bmc_decoder_ddr2 - Behavioral 
-- Project Name: 
-- Target Devices: 
-- Tool versions: 
-- Description: 
--
-- Dependencies: 
--
-- Revision: 
-- Revision 0.01 - File Created
-- Additional Comments: 
--
----------------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.all;

-- Uncomment the following library declaration if using
-- arithmetic functions with Signed or Unsigned values
use IEEE.NUMERIC_STD.all;

-- Uncomment the following library declaration if instantiating
-- any Xilinx primitives in this code.
--library UNISIM;
--use UNISIM.VComponents.all;
entity bmc_decoder_ddr2 is
  port (
    -- Inputs
    d_in     : in std_logic_vector (1 downto 0);
    clk      : in std_logic;  -- Some clock that is preferably 4x the datarate clock
    -- We will sync up to this clock. Since we are getting
    -- a ddr input we are at about 8x the input clock.
    enable_n : in std_logic;            -- Low Active Enable
    rst_n    : in std_logic;            -- Low Active Reset

    -- Outputs
    data       : out std_logic;
    data_ready : out std_logic;
    dclk_en       : out std_logic
    );
end bmc_decoder_ddr2;

architecture Behavioral of bmc_decoder_ddr2 is
  -- Edge detection signals
  signal d_buffer         : std_logic_vector (2 downto 0) := "000";  -- Holds buffer for transition detection oldest sample -> newest sample
  signal transition       : std_logic                     := '0';  -- Raise high when a transition is detected.
  signal transition_len   : unsigned (3 downto 0)         := x"0";  -- Length between transitions.
  signal transition_len_p : unsigned (3 downto 0)         := x"0";

  -- Clock Recovery
  signal recovered_clk : std_logic := '0';

  signal transitions_flag : std_logic := '0';

  signal transition_case1: boolean;
  signal transition_case2: boolean;
  signal half_period_cmp : boolean;
  signal full_period_cmp : boolean;

begin
  -- Buffer the recovered clock to an output
  dclk_en <= recovered_clk;

  -- [OBSOLETE] Since finding the period automatically is no longer happening
  -- The following code is obsolete, but exists to maintain
  -- compatability with a few of my modules. Should be removed in future
  -- revisions. - Chris W.
  data_ready <= '1';                -- Always stable. No need to check this any
                                    -- more
  -- [END OBSOLETE]
  
  -- Case where there is a transition on d_in
  transition_case1 <= (d_buffer(1 downto 0) = "01") or
                      (d_buffer(1 downto 0) = "10");
  -- Case where there was a transition between the last d_in and the new d_in
  transition_case2 <= (d_buffer(2 downto 1) = "01") or
                      (d_buffer(2 downto 1) = "10");

  -- Create the d_buffer.
  shift_buffer: process (clk)
  begin
    if rising_edge(clk) then
      d_buffer <= d_buffer(0) & d_in;
    end if;
  end process;
  
  -- Find transitions on the incoming data line.
  -- Uses a DDR data input (d_in)
  find_transitions : process (clk)
  begin
    --if (rst_n = '0') then
    --  transition_len   <= x"0";
    --  transition_len_p <= x"0";
      
    if rising_edge(clk) then
--      else        

        if transition_case1 then
          -- (d_buffer(1 downto 0) = "01") or (d_buffer(1 downto 0) = "10"
          transition_len   <= x"1";    -- New inputs are different. Start new
                                        -- transition length at 1 to account
                                        -- for this
          transition_len_p <= transition_len + x"1";  -- Add an additional 1
                                                       -- to the previous
                                                       -- transition_length.
          transition       <= '1';
        elsif transition_case2 then
        --  (d_buffer(2 downto 1) = "01") or (d_buffer(2 downto 1) = "10");
          -- If the transition occurs in the second part of the message we already added to transition_len from 
          -- the previous cycle. Clear transition_len.
          transition       <= '1';
          transition_len   <= x"2";    -- Add two because of doubel data rate
                                        -- (both new inputs are the same)
          transition_len_p <= transition_len;  -- No additional counts seen.
        else
          -- If we have not detected a transition add 2 to transition_len, to account for both new bits.
          transition_len_p <= transition_len; 
          transition_len   <= transition_len + x"2";  -- We are DDR so add 2
          transition       <= '0';
        end if;
  --    end if;
    end if;
  end process;  -- find_transitions

  half_period_cmp <= (transition_len_p = x"3") or
                     (transition_len_p = x"4") or
                     (transition_len_p = x"5");
  full_period_cmp <= (transition_len_p = x"7") or
                     (transition_len_p = x"8") or
                     (transition_len_p = x"9");
  -- Generate recovered clock and dataline from the preset period
  -- information
  recover_clk_and_data : process (clk, rst_n)
  begin
    if rst_n = '0' then
      data             <= '0';
      transitions_flag <= '0';
      recovered_clk    <= '0';

    elsif rising_edge(clk) then
--      else
        if transition = '1' then
          -- Half period count detected
--          if (transition_len_p = x"3") or (transition_len_p = x"4") or (transition_len_p = x"5") then
          if half_period_cmp then
            if (transitions_flag = '1') then
              -- If thi is the second half period count, output a '1' on data
              data             <= '1';
              transitions_flag <= '0';
              recovered_clk    <= '1';
            else
              -- If this is the first half period set the transistions flag.
              -- Takes 2 half period transitions for a '1' to occur
              recovered_clk    <= '0';
              transitions_flag <= '1';
            end if;
--          elsif (transition_len_p = x"7") or (transition_len_p = x"8") or (transition_len_p = x"9") then
            elsif full_period_cmp then
            -- Full period count detected output a '0' on data
            data             <= '0';
            recovered_clk    <= '1';
            transitions_flag <= '0';
          else
            transitions_flag <= '0';
          end if;
        else
          if (transition_len_p = x"3") or (transition_len_p = x"4") or (transition_len_p = x"5") then
          
            -- Toggle clock back after every data transitions.
            recovered_clk <= '0';  -- Not really a clock, more of a strobe.
          end if;
        end if;
  --    end if;
    end if;
  end process;  -- recover_clk_and_data

end Behavioral;
