----------------------------------------------------------------------------------
-- Company: Boston University EDF
-- Engineer: Christopher Woodall
-- 
-- Create Date:    11:31:11 07/26/2012 
-- Design Name: 
-- Module Name:    ttc_chb - Behavioral 
-- Project Name: 
-- Target Devices: 
-- Tool versions: 
-- Description: 
--
-- Dependencies: 
--
-- Revision: 
-- Revision 0.01 - File Created
-- Additional Comments: 
--
----------------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.all;
use IEEE.std_logic_misc.all;

-- Uncomment the following library declaration if using
-- arithmetic functions with Signed or Unsigned values
--use IEEE.NUMERIC_STD.ALL;

-- Uncomment the following library declaration if instantiating
-- any Xilinx primitives in this code.
--library UNISIM;
--use UNISIM.VComponents.all;

entity ttc_chb is
  port (
    data           : in  std_logic;
    clk            : in  std_logic;
    clk_en         : in std_logic;
    rst_n          : in  std_logic;
    broadcast_data : out std_logic_vector (7 downto 0);
    broadcast_strobe : out std_logic;
    double_err     : out std_logic;
    single_err     : out std_logic;
    frame_err      : out std_logic
    );
end ttc_chb;

architecture Behavioral of ttc_chb is
  type chb_state is (IDLE, START, BROADCAST_MSG, CMD_MSG);

  signal broadcast_data_buf : std_logic_vector(7 downto 0) := x"00";
  signal broadcast_count    : integer                      := 0;
  signal broadcast_chck     : std_logic_vector(4 downto 0) := "00000";
  signal broadcast_syn      : std_logic_vector(4 downto 0) := "00000";

  signal state : chb_state := IDLE;
begin
  -- Convert to synchronous
  decode_data : process(clk)
  begin
    if (rst_n = '0') then
      broadcast_count <= 0;
      broadcast_chck  <= "00000";
      broadcast_syn   <= "00000";
      broadcast_strobe <= '0';
    elsif rising_edge(clk) then
--      else
      if (clk_en = '1') then
          case state is
            when IDLE =>
              broadcast_strobe <='0';
              -- If a '0' comes in we got a start bit, so change state to START
              if data = '0' then
                state <= START;
              end if;
            when START =>
              -- There are two options, one of which is ignored.
              -- When FMT bit is '1' a individually addressed message is being received (IGNORE FOR NOW)
              -- WHen FMT bit is '0' a BROADCAST_MSG is being received. Go to the BROADCAST_MSG state
              if data = '1' then
                state <= IDLE;
              elsif data = '0' then
                state <= BROADCAST_MSG;
              end if;
            when BROADCAST_MSG =>
              -- A broadcast message has 8 data bits
              -- and 5 error detection bits (hamming code)
              if broadcast_count < 8 then
                broadcast_data_buf <= broadcast_data_buf (6 downto 0) & data;
                broadcast_count    <= broadcast_count + 1;
              elsif broadcast_count < 13 then
                broadcast_count <= broadcast_count + 1;
                broadcast_chck  <= broadcast_chck (3 downto 0) & data;
              else
                -- End of the frame, looking for stop bit ('1')
                -- Continue the same, but raise a frame error
                -- If a '0' is received there is a frame error. 
                frame_err <= not data;

                -- Check Hamming Code
                -- Hamming Correction and error logic from Wu's code
                -- TTC Hamming encoding

                --         hmg[0] = d[0]^d[1]^d[2]^d[3];
                -- hmg[1] = d[0]^d[4]^d[5]^d[6];
                -- hmg[2] = d[1]^d[2]^d[4]^d[5]^d[7];
                -- hmg[3] = d[1]^d[3]^d[4]^d[6]^d[7];
                -- hmg[4] = d[0]^d[2]^d[3]^d[5]^d[6]^d[7];              
                broadcast_syn(0) <= broadcast_chck(0) xor broadcast_data_buf(0) xor
                                    broadcast_data_buf(1) xor broadcast_data_buf(2) xor
                                    broadcast_data_buf(3);
                broadcast_syn(1) <= broadcast_chck(1) xor broadcast_data_buf(0) xor
                                    broadcast_data_buf(4) xor broadcast_data_buf(5) xor
                                    broadcast_data_buf(6);
                broadcast_syn(2) <= broadcast_chck(2) xor broadcast_data_buf(1) xor
                                    broadcast_data_buf(2) xor broadcast_data_buf(4) xor
                                    broadcast_data_buf(5) xor broadcast_data_buf(7);
                broadcast_syn(3) <= broadcast_chck(3) xor broadcast_data_buf(1) xor
                                    broadcast_data_buf(3) xor broadcast_data_buf(4) xor
                                    broadcast_data_buf(6) xor broadcast_data_buf(7);
                broadcast_syn(4) <= broadcast_chck(4) xor broadcast_data_buf(0) xor
                                    broadcast_data_buf(2) xor broadcast_data_buf(3) xor
                                    broadcast_data_buf(5) xor broadcast_data_buf(6) xor
                                    broadcast_data_buf(7);
  
                -- Detect if there is a single bit error
                single_err <= xor_reduce(broadcast_syn) and data;  -- data == not frame_err
  
                -- Detect if there has been a double bit error.
                if ((or_reduce(broadcast_syn) = '1') and (xor_reduce(broadcast_syn) = '0')) or (data = '0') then
                  double_err <= '1';
                else
                  double_err <= '0';
                end if;
  
                -- Make corrections for single bit errors!
                if broadcast_syn(3 downto 0) = x"c" then
                  broadcast_data(7) <= not broadcast_data_buf(7);
                else
                  broadcast_data(7) <= broadcast_data_buf(7);
                end if;
  
                if broadcast_syn(3 downto 0) = x"a" then
                  broadcast_data(6) <= not broadcast_data_buf(6);
                else
                  broadcast_data(6) <= broadcast_data_buf(6);
                end if;
  
                if broadcast_syn(3 downto 0) = x"6" then
                  broadcast_data(5) <= not broadcast_data_buf(5);
                else
                  broadcast_data(5) <= broadcast_data_buf(5);
                end if;
  
                if broadcast_syn(3 downto 0) = x"e" then
                  broadcast_data(4) <= not broadcast_data_buf(4);
                else
                  broadcast_data(4) <= broadcast_data_buf(4);
                end if;
  
                if broadcast_syn(3 downto 0) = x"9" then
                  broadcast_data(3) <= not broadcast_data_buf(3);
                else
                  broadcast_data(3) <= broadcast_data_buf(3);
                end if;
  
                if broadcast_syn(3 downto 0) = x"5" then
                  broadcast_data(2) <= not broadcast_data_buf(2);
                else
                  broadcast_data(2) <= broadcast_data_buf(2);
                end if;
  
                if broadcast_syn(3 downto 0) = x"d" then
                  broadcast_data(1) <= not broadcast_data_buf(1);
                else
                  broadcast_data(1) <= broadcast_data_buf(1);
                end if;
  
                if broadcast_syn(3 downto 0) = x"3" then
                  broadcast_data(0) <= not broadcast_data_buf(0);
                else
                  broadcast_data(0) <= broadcast_data_buf(0);
                end if;
  
                -- Reset
                broadcast_count    <= 0;      --clear counter
                broadcast_data_buf <= x"00";  -- clear broadcast_buffe
                broadcast_strobe <= '1';
                state              <= IDLE;
              end if;
            when others =>
              state <= IDLE;
          end case;
        end if;
      
    end if;
  end process;

end Behavioral;
