--
-- enforce trigger rules: no more than n triggers in m clocks
--

library IEEE;
use IEEE.STD_LOGIC_1164.all;
use IEEE.STD_LOGIC_ARITH.all;
use IEEE.STD_LOGIC_UNSIGNED.all;
--
library unisim;
use unisim.vcomponents.all;

entity rule is

  generic (
    max_bx  : integer := 1;
    history : integer := 16);

  port (
    clk      : in  std_logic;
    rst_n    : in  std_logic;
    rule     : in  std_logic_vector(7 downto 0);
    trig_in  : in  std_logic;
    trig_out : out std_logic;
    violate  : out std_logic
    );

end rule;


architecture arch of rule is

  signal trig_sr  : std_logic_vector((history-1) downto 0);
  signal count    : std_logic_vector(7 downto 0);
  signal inhibit  : std_logic;
  signal trig_reg : std_logic;

begin  -- arch

  process (clk, rst_n)
  begin  -- process
    if rst_n = '0' then                 -- asynchronous reset (active low)

      trig_sr <= (others => '0');
      count   <= (others => '0');

    elsif clk'event and clk = '1' then  -- rising clock edge

      -- trigger history shift register:  inhibit transfer from 1->2 if rule violated
      trig_sr <= trig_sr( (history-2) downto 2) & ((not inhibit) and trig_sr(1)) & trig_sr(0) & trig_in;

      inhibit <= '0';

      -- incoming trigger, no out-going one: count up for rule
      if (trig_sr(0) = '1') and (trig_sr( conv_integer( rule)) = '0') then
        -- if we're already at the threshold inhibit the count
        if count = max_bx then
          inhibit <= '1';
        else
          count   <= count + 1;
        end if;
      end if;

      -- outgoing trigger, no incoming one: count down for rule
      if (trig_sr(0) = '0') and (trig_sr( conv_integer( rule)) = '1') then
        count <= count - 1;
      end if;

      -- generate output tirgger if not inhibited
      trig_out   <= '0';
      if (inhibit = '0') and (trig_sr(1) = '1') then
        trig_out <= '1';
      end if;

      violate <= inhibit;

    end if;
  end process;



end arch;
