--
-- trigger generator
-- random or periodic triggers
-- enforce up to 4 trigger rules
-- blanking period around orbit signal
--

library IEEE;
use IEEE.STD_LOGIC_1164.all;
use IEEE.STD_LOGIC_ARITH.all;
use IEEE.STD_LOGIC_UNSIGNED.all;
--
library unisim;
use unisim.vcomponents.all;

entity tts_control is

  port (
    parm    : in  std_logic_vector(255 downto 0);  -- Parameters 
    orb_clk : in  std_logic_vector(3 downto 0);
    clk80   : in  std_logic;
    tts_a   : in  std_logic_vector(3 downto 0);
    cr      : in  std_logic_vector(7 downto 0);  -- Control Register
    ar      : in  std_logic_vector(7 downto 0);  -- Action Register
    sr      : out std_logic_vector(7 downto 0);  -- Status Register
    dr      : out std_logic_vector(7 downto 0);  -- Data Register
    bx_count : in std_logic_vector(11 downto 0);
    nim_l1a : out std_logic;
    nim_ttc : out std_logic;
    test    : out std_logic_vector(1 downto 0);
    ecl_bc0 : out std_logic);

end tts_control;


architecture arch of tts_control is

  component rule
    generic (
      max_bx   :     integer;
      history  :     integer);
    port (
      clk      : in  std_logic;
      rst_n    : in  std_logic;
      rule     : in  std_logic_vector(7 downto 0);
      trig_in  : in  std_logic;
      trig_out : out std_logic;
      violate  : out std_logic);
  end component;

  component urand
    port (
      clk   : in  std_logic;
      rst_n : in  std_logic;
      u     : out std_logic_vector(31 downto 0);
      dv    : out std_logic);
  end component;

  component synch
    port (
      clk, rst, s : in  std_logic;
      y           : out std_logic);
  end component;

  component fifo3
    port (
      din    : in  std_logic_vector(63 downto 0);
      rd_clk : in  std_logic;
      rd_en  : in  std_logic;
      rst    : in  std_logic;
      wr_clk : in  std_logic;
      wr_en  : in  std_logic;
      dout   : out std_logic_vector(63 downto 0);
      empty  : out std_logic;
      full   : out std_logic);
  end component;

  signal sys_clk : std_logic;
  signal bc0_clk : std_logic;
  signal clk_sel : std_logic_vector(1 downto 0);

--  signal bx_count    : std_logic_vector(11 downto 0);  -- BX counter
  signal orbit_count : std_logic_vector(15 downto 0);  -- orbit counter
  signal l1a_count   : std_logic_vector(7 downto 0);  -- count of L1A this turn
  signal next_l1a    : std_logic_vector(11 downto 0);  -- count down to next L1A

  signal trigger_count     : std_logic_vector(31 downto 0);
  signal trigger_count_reg : std_logic_vector(31 downto 0);

  -- parameters
  signal orbit_length   : std_logic_vector(11 downto 0);
  signal trig_start     : std_logic_vector(11 downto 0);
  signal orbit_max      : std_logic_vector(15 downto 0);
  signal l1a_spacing    : std_logic_vector(11 downto 0);
  signal l1a_per_orbit  : std_logic_vector(7 downto 0);
  signal repeat_period  : std_logic_vector(15 downto 0);
  signal set_bp_delay   : std_logic_vector(7 downto 0);  -- bp delay setting
  signal bp_sample_mask : std_logic_vector(7 downto 0);

  -- trigger rules: minimum spacing for 1..4 triggers
  signal rule_1 : std_logic_vector(7 downto 0);
  signal rule_2 : std_logic_vector(7 downto 0);
  signal rule_3 : std_logic_vector(7 downto 0);
  signal rule_4 : std_logic_vector(7 downto 0);

  signal violate_1 : std_logic_vector(7 downto 0);  -- count times rule violated
  signal violate_2 : std_logic_vector(7 downto 0);  -- count times rule violated
  signal violate_3 : std_logic_vector(7 downto 0);  -- count times rule violated
  signal violate_4 : std_logic_vector(7 downto 0);  -- count times rule violated

  signal v1, v2, v3, v4 : std_logic;

  signal action : std_logic_vector(7 downto 0);

  signal rst_n, rst : std_logic;

  signal ena_l1a      : std_logic;      -- L1A enabled (synch to orbit start)
  signal req_l1a      : std_logic;      -- L1A requested from software
  signal live_l1a     : std_logic;      -- L1A active now (this part of orbit)
  signal want_l1a     : std_logic;      -- L1A scheduled (can be disabled by
                                        -- backpressure) 
  signal mux_l1a      : std_logic;      -- either programmed or random L1A
  signal inhibit      : std_logic;      -- inhibit due to trigger rules
  signal issue_l1a    : std_logic;      -- L1A requested (after backpressure)
  signal real_trigger : std_logic;      -- L1A issued (after trigger rules)

  signal trig_out_1 : std_logic;        -- interconnect between rules
  signal trig_out_2 : std_logic;        -- interconnect between rules
  signal trig_out_3 : std_logic;        -- interconnect between rules  

  signal do_repeat : std_logic;         -- repeat sequence indefinitely
  signal do_burst  : std_logic;         -- more than 1 trigger per orbit
  signal do_random : std_logic;         -- enable random triggers

  signal orbit_stretch : std_logic_vector(3 downto 0);
  signal orbit_int     : std_logic;

  signal tts_0, tts_1, tts_2, tts_3 : std_logic_vector(3 downto 0);
  signal tts_new                    : std_logic_vector(63 downto 0);
  signal tts_changed                : std_logic;

  signal fifo_out   : std_logic_vector(63 downto 0);
  signal fifo_byte  : std_logic_vector(3 downto 0);
  signal fifo_count : std_logic_vector(9 downto 0);

  signal urand32, rand32 : std_logic_vector(31 downto 0);
  signal counter_byte    : std_logic_vector(1 downto 0);
  signal counter_mux     : std_logic_vector(7 downto 0);

  signal rate_set      : std_logic_vector(15 downto 0);
  signal rand_trig_req : std_logic;

  signal rand_dv : std_logic;

  signal time_stamp : std_logic_vector(59 downto 0);

  signal backpressure : std_logic;      -- enable backpressure signals
  signal bp_stop_in   : std_logic;      -- request to stop L1A from TTS
  signal bp_delay     : std_logic_vector(255 downto 0);  -- backpressure delay

  signal saw_ofw : std_logic;           -- OFW is active
  signal saw_bsy : std_logic;           -- BSY is active
  signal saw_syn : std_logic;           -- saw SYN

  signal ttc_cmd_prescale : std_logic_vector(15 downto 0);
  signal ttc_bx_num       : std_logic_vector(11 downto 0);

  signal ena_ttc : std_logic;

  -- functions to access a byte in a large vector
  function get_bits ( v : std_logic_vector; lo : integer; len : integer)
    return std_logic_vector is
  begin
    return v( lo+len-1 downto lo);
  end function get_bits;

  -- note:  BCZ TTS command comes ~ 52 clocks after this signal
  constant front_porch : integer := 250;  -- empty bunches after orbit signal
  constant back_porch  : integer := 500;  -- empty bunches before orbit signal

  signal orbit_blank_disable : std_logic;

begin  -- arch

  rst   <= action(7);
  rst_n <= not rst;

  sys_clk <= orb_clk(0);

  -- get parameters
  orbit_length  <= get_bits( parm, 0, 12);    -- parm 0, 1
  trig_start    <= get_bits( parm, 16, 12);   -- parm 2, 3
  orbit_max     <= get_bits( parm, 32, 16);   -- parm 4, 5
  l1a_spacing   <= get_bits( parm, 48, 12);   -- parm 6, 7
  l1a_per_orbit <= get_bits( parm, 64, 8);    -- parm 8
  repeat_period <= get_bits( parm, 72, 16);   -- parm 9, 10
  fifo_byte     <= get_bits( parm, 88, 4);    -- parm 11
  counter_byte  <= get_bits( parm, 96, 2);    -- parm 12
  rate_set      <= get_bits( parm, 104, 16);  -- param 13, 14

  rule_1 <= get_bits( parm, 16*8, 8);   -- param 16
  rule_2 <= get_bits( parm, 17*8, 8);   -- param 17
  rule_3 <= get_bits( parm, 18*8, 8);   -- param 18
  rule_4 <= get_bits( parm, 19*8, 8);   -- param 19

  clk_sel        <= get_bits( parm, 20*8, 2);  -- param 20
  set_bp_delay   <= get_bits( parm, 21*8, 8);  -- param 21
  bp_sample_mask <= get_bits( parm, 22*8, 8);  -- param 22

  ttc_bx_num       <= get_bits( parm, 23*8, 12);  -- param 23, 24
  ttc_cmd_prescale <= get_bits( parm, 25*8, 16);  -- param 25, 26

  -- get control reg bits
  do_burst            <= cr(0);
  do_repeat           <= cr(1);
  backpressure        <= cr(2);
  do_random           <= cr(3);
  orbit_blank_disable <= cr(4);

  test(0) <= tts_changed;
  test(1) <= '0';

  -- mux fifo output to data register
  with fifo_byte select
    dr <=
    fifo_out(7 downto 0)              when "0000",
    fifo_out(15 downto 8)             when "0001",
    fifo_out(23 downto 16)            when "0010",
    fifo_out(31 downto 24)            when "0011",
    fifo_out(39 downto 32)            when "0100",
    fifo_out(47 downto 40)            when "0101",
    fifo_out(55 downto 48)            when "0110",
    fifo_out(63 downto 56)            when "0111",
    fifo_count(7 downto 0)            when "1000",
    "000000" & fifo_count(9 downto 8) when "1001",
    trigger_count_reg(7 downto 0)     when "1010",
    trigger_count_reg(15 downto 8)    when "1011",
    trigger_count_reg(23 downto 16)   when "1100",
    trigger_count_reg(31 downto 24)   when "1101",
    counter_mux                       when others;

  with counter_byte select
    counter_mux <=
    violate_1 when "00",
    violate_2 when "01",
    violate_3 when "10",
    violate_4 when others;

  bc0_clk <= orb_clk(0);

  process (bc0_clk, rst_n)
  begin  -- process
    if rst_n = '0' then                 -- asynchronous reset (active low)

    elsif bc0_clk'event and bc0_clk = '1' then
      ecl_bc0 <= orbit_int;
    end if;
  end process;

  process (sys_clk, rst_n)
  begin  -- process
    if rst_n = '0' then                 -- asynchronous reset (active low)

      saw_bsy <= '0';
      saw_ofw <= '0';
      saw_syn <= '0';

      ena_l1a       <= '0';
      live_l1a      <= '0';
      orbit_count   <= (others => '0');
--      bx_count      <= (others => '0');
      l1a_count     <= (others => '0');
      next_l1a      <= (others => '0');
      fifo_count    <= (others => '0');
      time_stamp    <= (others => '0');
      trigger_count <= (others => '0');

      violate_1 <= (others => '0');
      violate_2 <= (others => '0');
      violate_3 <= (others => '0');
      violate_4 <= (others => '0');

    elsif sys_clk'event and sys_clk = '1' then  -- rising clock edge

      time_stamp <= time_stamp + 1;

      orbit_stretch( 3 downto 1) <= orbit_stretch(2 downto 0);

      orbit_stretch(0) <= '1';
      want_l1a         <= '0';

      orbit_int <= orbit_stretch(0) and orbit_stretch(1) and
                   orbit_stretch(2);

      -- BX number to send TTC command
      nim_ttc     <= '0';
      if bx_count = ttc_bx_num then
        if ena_ttc = '1' then
          ena_ttc <= '0';
          nim_ttc <= '1';
        end if;
      end if;

      -- BX number to start trigger stream
      if bx_count = trig_start then
        if ena_l1a = '1' then
          if do_burst = '1' then        -- want a burst?
            live_l1a <= '1';
          end if;
          want_l1a   <= '1';            -- make first L1A
          next_l1a   <= (others => '0');
          l1a_count  <= (0      => '1', others => '0');
        end if;
      end if;

      -- request L1A from software
      if action(0) = '1' then
        req_l1a <= '1';

        saw_bsy       <= '0';
        saw_ofw       <= '0';
        saw_syn       <= '0';
        trigger_count <= (others => '0');

      end if;

      if live_l1a = '1' then                 -- are L1A live?
        if next_l1a = l1a_spacing then       -- time for next one?
          next_l1a    <= (others => '0');    -- yes, reset count
          want_l1a    <= '1';                -- output L1A pulse
          if l1a_count = l1a_per_orbit then  -- any L1A left this orbit?
            live_l1a  <= '0';                -- nope, we're done
          else
            l1a_count <= l1a_count + 1;      -- yes, keep counting
          end if;
        else
          next_l1a    <= next_l1a + 1;
        end if;
      end if;

      -- start of new orbit
      if bx_count = orbit_length then

--        bx_count         <= (others => '0');  -- reset bunch count
        orbit_stretch(0) <= '0';

        live_l1a <= '0';

        -- synchronize L1A request to orbit
        if req_l1a = '1' then
          orbit_count <= (others => '0');
          ena_l1a     <= '1';
          ena_ttc     <= '1';
          req_l1a     <= '0';
        else
          orbit_count <= orbit_count + 1;
        end if;



        -- disable L1A if we reach programmed orbit count
        if orbit_count = orbit_max then
          ena_l1a <= '0';
        end if;

        -- generate a new request if repeats are needed
        if do_repeat = '1' then
          if orbit_count = repeat_period then
            req_l1a <= '1';
          end if;
        end if;

--      else
--        bx_count <= bx_count + 1;
      end if;

      -- sample TTS, write FIFO if any change
      tts_3 <= tts_2;
      tts_2 <= tts_1;
      tts_1 <= tts_0;
      tts_0 <= tts_a;

      -- any change must persist
      tts_changed <= '0';
      if (tts_2 /= tts_1) and (tts_0 = tts_1) then

        tts_new    <= time_stamp & tts_0;
        tts_changed <= '1';
        fifo_count  <= fifo_count + 1;

--        if tts_0 = "0001" then
        if tts_0 = "0100" then
          saw_bsy <= '1';
        else
          saw_bsy <= '0';
        end if;

--        if tts_0 = "0100" then
        if tts_0 = "0001" then
          saw_ofw <= '1';
        else
          saw_ofw <= '0';
        end if;

--        if tts_0 = "1000" then
        if tts_0 = "0010" then
          saw_syn <= '1';
        else
          saw_syn <= '0';
        end if;

      end if;

      -- delay backpressure as programmed
      if (saw_bsy = '1') or (saw_syn = '1') or (saw_ofw = '1') then
        bp_stop_in <= '1';
      else
        bp_stop_in <= '0';
      end if;

      -- advance bp_delay shift register when bx_count bits
      -- selected by bp_sample_mask are zero
      if (bx_count(7 downto 0) and bp_sample_mask) = X"00" then
        bp_delay <= bp_delay(254 downto 0) & bp_stop_in;
      end if;

      -- deal with backpressure
      if backpressure = '1' then

        -- disable L1A now if SYN or BSY or OFW
        if bp_delay( conv_integer(set_bp_delay)) = '1' then
          issue_l1a <= '0';
        else
          issue_l1a <= mux_l1a;
        end if;
      else
        issue_l1a   <= mux_l1a;
      end if;

      -- deal with orbit blanking
      if (orbit_blank_disable = '0') and
        ((bx_count < front_porch) or (bx_count > (3563-back_porch))) then
        nim_l1a <= '1';
      else
        nim_l1a <= not real_trigger;
      end if;

      -- count triggers
      if real_trigger = '1' then
        trigger_count <= trigger_count + 1;
      end if;

      -- latch count on action 2
      if action(2) = '1' then
        trigger_count_reg <= trigger_count;
      end if;

      -- random number generator
      rand_trig_req     <= '0';
      if rand_dv = '1' then
        rand32          <= urand32;
        if rand32(15 downto 0) < rate_set then
          rand_trig_req <= '1';
        end if;
      end if;

      -- mux programmed triggers
      if do_random = '1' then
        mux_l1a <= rand_trig_req;
      else
        mux_l1a <= want_l1a;
      end if;

      -- count rule violations
      if v1 = '1' then
        violate_1 <= violate_1 + 1;
      end if;

      if v2 = '1' then
        violate_2 <= violate_2 + 1;
      end if;

      if v3 = '1' then
        violate_3 <= violate_3 + 1;
      end if;

      if v4 = '1' then
        violate_4 <= violate_4 + 1;
      end if;

    end if;
  end process;

  l1   : for i in 0 to 7 generate
    s1 : synch
      port map (
        clk => sys_clk,
        rst => '1',
        s   => ar(i),
        y   => action(i));
  end generate l1;

  f1 : fifo3
    port map (
      din    => tts_new,
      rd_clk => sys_clk,
      rd_en  => action(1),
      rst    => rst,
      wr_clk => sys_clk,
      wr_en  => tts_changed,
      dout   => fifo_out,
      empty  => sr(0),
      full   => sr(1));

  ran1 : urand
    port map (
      clk   => clk80,
      rst_n => rst_n,
      u     => urand32,
      dv    => rand_dv);

-- trigger rules
  r1 : rule
    generic map (
      history  => 256,
      max_bx   => 1)
    port map (
      clk      => sys_clk,
      rst_n    => rst_n,
      rule     => rule_1,
      trig_in  => issue_l1a,
      trig_out => trig_out_1,
      violate  => v1);

  r2 : rule
    generic map (
      history  => 256,
      max_bx   => 2)
    port map (
      clk      => sys_clk,
      rst_n    => rst_n,
      rule     => rule_2,
      trig_in  => trig_out_1,
      trig_out => trig_out_2,
      violate  => v2);

  r3 : rule
    generic map (
      history  => 256,
      max_bx   => 3)
    port map (
      clk      => sys_clk,
      rst_n    => rst_n,
      rule     => rule_3,
      trig_in  => trig_out_2,
      trig_out => trig_out_3,
      violate  => v3);

  r4 : rule
    generic map (
      history  => 256,
      max_bx   => 4)
    port map (
      clk      => sys_clk,
      rst_n    => rst_n,
      rule     => rule_4,
      trig_in  => trig_out_3,
      trig_out => real_trigger,
      violate  => v4);

end arch;
